#include <apr_strings.h>
#include <ctype.h>
#include <dmlite/c/catalog.h>
#include <dmlite/c/checksums.h>
#include <httpd.h>
#include <http_log.h>
#include <regex.h>
#include <string.h>
#include "mod_lcgdm_disk.h"

#define RFC3230_REGEXP "^([[:alnum:]]+)(;q=[[:digit:]]*)?([,]?)"

static char* dav_disk_strcpy_lower(char *out, const char *in, size_t max)
{
    size_t i;
    for (i = 0; in[i] && i < max; ++i)
        out[i] = tolower(in[i]);
    return out;
}

/**
 * Parses p in order to find the next digest algorithm
 * We don't care much about priorities, so give it as we find them
 * @return 0 when there are no more digest available
 */
int dav_disk_next_digest(const char **want_digest, char *output, size_t outsize)
{
    (void) outsize;

    regex_t regex;
    regmatch_t matches[3];

    if (regcomp(&regex, RFC3230_REGEXP, REG_EXTENDED) != 0)
        abort();

    while (isspace(**want_digest))
        ++(*want_digest);

    if (regexec(&regex, *want_digest, 3, matches, 0) == 0) {
        size_t digest_len = matches[1].rm_eo - matches[1].rm_so;
        const char *digest = (*want_digest) + matches[1].rm_so;

        dav_disk_strcpy_lower(output, digest, digest_len);
        output[digest_len] = '\0';

        *want_digest += matches[0].rm_eo;

        return 1;
    }

    return 0;
}

/**
 * Queries the nameserver. If it can not get the checksum back,
 * it computes it, and tries to set it back in the nameserver
 * @return 0 If the checksum was not found, and can not be calculated
 */
static int cached_checksum(request_rec *r, const dav_resource_private *info,
        const char* digest_name, char *digest_buffer, size_t bsize)
{
    char *digest_xattr;

    // Build digest xattr
    digest_xattr = apr_pstrcat(r->pool, "checksum.", digest_name, NULL );

    // Try from namespace
    dmlite_xstat xstat;

    xstat.extra = dmlite_any_dict_new();
    dmlite_statx(info->ctx, info->namespace_path, &xstat);
    dmlite_any* digest_value = dmlite_any_dict_get(xstat.extra, digest_xattr);

    if (digest_value) {
        dmlite_any_to_string(digest_value, digest_buffer, bsize);
        dmlite_any_free(digest_value);
        dmlite_any_dict_free(xstat.extra);
        ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
                "Digest hit in the namespace for %s! %s = %s", r->uri,
                digest_name, digest_buffer);
        return 1;
    }
    else {
        ap_log_rerror(APLOG_MARK, APLOG_WARNING, 0, r,
                "Digest miss in the namespace for %s (%s)", r->uri,
                digest_name);
    }

    // Calculate
    int got_one = 0;

    if (strcmp("md5", digest_name) == 0)
        got_one = (dmlite_checksum_md5(info->fd, 0, 0, digest_buffer, bsize)
                == 0);
    else if (strcmp("adler32", digest_name) == 0)
        got_one = (dmlite_checksum_adler32(info->fd, 0, 0, digest_buffer, bsize)
                == 0);
    else if (strcmp("crc32", digest_name) == 0)
        got_one = (dmlite_checksum_crc32(info->fd, 0, 0, digest_buffer, bsize)
                == 0);

    // Set it in the namespace
    if (got_one) {
        digest_value = dmlite_any_new_string(digest_buffer);
        dmlite_any_dict_insert(xstat.extra, digest_xattr, digest_value);
        dmlite_any_free(digest_value);

        if (dmlite_update_xattr(info->ctx, info->namespace_path, xstat.extra)
                == 0) {
            ap_log_rerror(APLOG_MARK, APLOG_INFO, 0, r,
                    "Digest set in the namespace in the namespace for %s (%s = %s)",
                    r->uri, digest_name, digest_buffer);
        }
        else {
            ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
                    "Could not set the digest in the namespace for %s", r->uri);
        }
    }
    else {
        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
                "Could not calculate the digest %s for %s", digest_name,
                r->uri);
    }

    dmlite_any_dict_free(xstat.extra);

    return got_one;
}

int dav_disk_digest_header(request_rec *r, const dav_resource *resource,
        char* output, size_t outsize)
{
    const char *want_digest;
    char digest_name[32], digest[64];
    int has_digest;

    want_digest = apr_table_get(r->headers_in, "Want-Digest");

    if (!resource->info->fd || !want_digest)
        return 0;

    has_digest = 0;
    while (dav_disk_next_digest(&want_digest, digest_name, sizeof(digest_name))) {
        if (cached_checksum(r, resource->info, digest_name, digest, sizeof(digest))) {
            size_t len;
            len = snprintf(output, outsize, "%s=%s,", digest_name, digest);
            output += len;
            outsize -= len;
            has_digest = 1;
            dmlite_fseek(resource->info->fd, 0, SEEK_SET);
        }
    }

    --output;
    *output = '\0';

    return has_digest;
}
