/**
 * @file   copy.c
 * @brief  Third-party copy implementation.
 * @author Alejandro Álvarez Ayllón <aalvarez@cern.ch>
 */
#include <apr_buckets.h>
#include <apr_strings.h>
#include <ctype.h>
#include <dmlite/c/io.h>
#include <httpd.h>
#include <http_log.h>
#include <http_protocol.h>
#include <stdarg.h>
#include <sys/stat.h>
#include <util_filter.h>

#include "delegation.h"
#include "mod_lcgdm_disk.h"
#include "../client/htext.h"
#include "../shared/utils.h"

/** Structure used to pass data along callbacks. */
typedef struct
{
    apr_bucket_brigade *brigade;
    ap_filter_t *output;
    request_rec *request;
    const char *remote;
} dav_disk_copy_data;

/* Custom I/O handler */
typedef struct
{
    off_t fsize;
    dmlite_context* context;
    const dmlite_any_dict* extra;
} dav_disk_open_params;

off_t dav_disk_fsize(const char *path, void *udata)
{
    (void) path;
    dav_disk_open_params* params = udata;
    return params->fsize;
}

void* dav_disk_fopen(const char *path, const char *mode, void *udata)
{
    dav_disk_open_params* params = udata;

    /* In fact, we only want read mode! */
    if (strchr(mode, 'w') != NULL || strchr(mode, '+') != NULL )
        return NULL ;

    return dmlite_fopen(params->context, path, O_RDONLY, params->extra);
}

int dav_disk_fclose(void *fd)
{
    return dmlite_fclose(fd);
}

size_t dav_disk_fread(void *buffer, size_t size, size_t n, void *fd)
{
    ssize_t nbytes = dmlite_fread(fd, buffer, size * n);
    if (nbytes <= 0)
        return 0;
    else
        return nbytes;
}

size_t dav_disk_fwrite(const void* buffer, size_t size, size_t n, void *fd)
{
    (void) buffer;
    (void) size;
    (void) n;
    (void) fd;
    return 0;
}

int dav_disk_fseek(void *fd, long offset, int whence)
{
    return dmlite_fseek(fd, offset, whence);
}

long dav_disk_ftell(void* fd)
{
    return dmlite_ftell(fd);
}

int dav_disk_feof(void* fd)
{
    return dmlite_feof(fd);
}

struct htext_io_handler dav_disk_io_handler = {
        dav_disk_fsize, dav_disk_fopen, dav_disk_fclose, dav_disk_fread,
        dav_disk_fwrite, dav_disk_fseek, dav_disk_ftell, dav_disk_feof };

/**
 * Called by htext when there is logging output
 * @param handle The handle that triggers the call
 * @param type   Type of logging data
 * @param msg    The message itself
 * @param size   The size of the message (it might not be NULL terminated)
 * @param ud     User defined data
 */
static void dav_disk_copy_log(htext_handle *handle, HTEXT_LOG_TYPE type,
        const char *msg, size_t size, void *ud)
{
    (void) handle;
    (void) size;

    dav_disk_copy_data *ddc = (dav_disk_copy_data*) ud;

    switch (type) {
        case HTEXT_LOG:
            ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, ddc->request, "%s", msg);
            break;
        default:
            break;
    }
}

/**
 * Generates and sends the performance clients following GridFTP format
 * http://www.ogf.org/documents/GFD.20.pdf (Appendix II), skipping the
 * code.
 */
static void dav_disk_send_performance_markers(dav_disk_copy_data *ddc, size_t n,
        size_t total[], size_t done[])
{
    (void) total;

    time_t timestamp = time(NULL );
    size_t i;

    for (i = 0; i < n; ++i) {
        apr_brigade_printf(ddc->brigade, ap_filter_flush, ddc->output,
                "Perf Marker\n"
                        "\tTimestamp: %ld\n"
                        "\tStripe Index: %u\n"
                        "\tStripe Bytes Transferred: %ld\n"
                        "\tTotal Stripe Count: %ld\n"
                        "End\n", timestamp, (unsigned) i, (long) done[i],
                (long) n);
    }
}

/**
 * Polls the handler, and sends feedback to the client
 */
static dav_error *dav_disk_pool_and_feedback(htext_handle *handle,
        dav_disk_copy_data *ddc)
{
    int wait, status;
    dav_error *error = NULL;
    const char *error_string;

    do {
        size_t *total, *done, n, i;
        size_t globalDone, globalTotal;

        status = htext_status(handle);
        switch (status) {
            case HTEXTS_SUCCEEDED:
            case HTEXTS_FAILED:
            case HTEXTS_ABORTED:
                wait = 0;
                break;
            default:
                /* In the first go we need to set the reply */
                if (ddc->request->status == 0) {
                    ddc->request->status = HTTP_ACCEPTED;
                    ap_set_content_type(ddc->request, "text/plain");
                }
                /* Print progress */
                htext_progress(handle, &n, &total, &done);
                globalDone = globalTotal = 0;
                for (i = 0; i < n; ++i) {
                    globalDone += done[i];
                    globalTotal += total[i];
                }

                dav_disk_send_performance_markers(ddc, n, total, done);

                if (ap_fflush(ddc->output, ddc->brigade) == APR_SUCCESS) {
                    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, ddc->request,
                            "COPY '%s' %lu/%lu", ddc->request->uri,
                            (unsigned long) globalDone,
                            (unsigned long) globalTotal);
                    wait = 1;
                    sleep(1);
                }
                else {
                    wait = 0;
                    htext_abort(handle);
                }
        }
    } while (wait);

    error_string = htext_error_string(handle);
    switch (status) {
        case HTEXTS_SUCCEEDED:
            ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, ddc->request,
                    "Remote copy finished successfully (%d): %s => %s",
                    htext_http_code(handle), ddc->request->uri, ddc->remote);
            /* 200 is OK, but 201 fits better */
            if (ddc->request->status == 0)
                ddc->request->status = HTTP_CREATED;
            apr_brigade_printf(ddc->brigade, ap_filter_flush, ddc->output,
                    "Success\n");
            break;
        case HTEXTS_FAILED:
            if (ddc->request->status == 0)
                error = dav_shared_new_error(ddc->request, NULL,
                        htext_http_code(handle),
                        "Failed: Remote copy failed with status code %d: %s\n",
                        htext_http_code(handle),
                        error_string ? error_string : "");
            else
                apr_brigade_printf(ddc->brigade, ap_filter_flush, ddc->output,
                        "Failed: Remote copy failed with status code %d: %s\n",
                        htext_http_code(handle),
                        error_string ? error_string : "");
            break;
        case HTEXTS_ABORTED:
            if (ddc->request->status == 0)
                error = dav_shared_new_error(ddc->request, NULL,
                        HTTP_INTERNAL_SERVER_ERROR, "Aborted");
            else
                apr_brigade_printf(ddc->brigade, ap_filter_flush, ddc->output,
                        "Aborted\n");
            break;
    }

    return error;
}

/* Exposed function. */
dav_error *dav_disk_remote_copy(const dav_resource *src, const char *remote)
{
    apr_uri_t remote_uri;
    dav_error *error = NULL;
    apr_bucket *bkt;
    char *uproxy;
    htext_handle *handle;
    dav_disk_copy_data ddc;
    dav_disk_open_params open_params;
    const char *client_name, *streams_header;
    int nstreams = 1;

    client_name =
            dmlite_get_security_context(src->info->ctx)->credentials->client_name;

    /* Parse URI */
    apr_uri_parse(src->pool, remote, &remote_uri);

    /* Check permissions */
    if (!(src->info->d_conf->flags & DAV_DISK_REMOTE_COPY)) {
        return dav_shared_new_error(src->info->request, NULL,
                HTTP_METHOD_NOT_ALLOWED,
                "This server does not allow remote copies");
    }

    /* Get delegation proxy */
    uproxy = dav_disk_get_proxy(src->info->request,
            src->info->d_conf->proxy_cache, client_name);

    if (!uproxy && apr_table_get(src->info->request->headers_in, "X-No-Delegate") == NULL) {
        if (src->info->copy_already_redirected) {
            return dav_shared_new_error(src->info->request, NULL,
                    419, // Need credentials
                    "Could not find a delegated proxy after an explicit request for delegation.\n"
                            "Probably your client do not support the header 'X-Delegate-To'?");
        }
        else if (src->info->d_conf->delegation_service != NULL ) {
            const char* redirect_with_delegation;
            redirect_with_delegation = apr_pstrcat(src->pool,
                    src->info->request->unparsed_uri, "&copyRedirected=1", NULL );

            apr_table_setn(src->info->request->err_headers_out, "X-Delegate-To",
                    src->info->d_conf->delegation_service);
            apr_table_setn(src->info->request->err_headers_out, "Location",
                    redirect_with_delegation);

            return dav_shared_new_error(src->info->request, NULL,
                    HTTP_MOVED_TEMPORARILY, "Could not find a delegated proxy.");
        }
        else {
            return dav_shared_new_error(src->info->request, NULL,
                    HTTP_INTERNAL_SERVER_ERROR,
                    "Could not find a delegated proxy, and there is no delegation endpoint configured.");
        }
    }

    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, src->info->request,
            "COPY request from '%s'", client_name);

    if (uproxy)
        ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, src->info->request,
                "Using delegation proxy '%s'", uproxy);
    else
        ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, src->info->request,
                "Not doing the delegation (X-No-Delegate header present)");

    /* Multiple streams? */
    streams_header = apr_table_get(src->info->request->headers_in,
            "X-Number-Of-Streams");
    if (streams_header) {
        long val = strtol(streams_header, NULL, 10);
        if (val >= 1 && val < 32)
            nstreams = (int) val;
    }

    /* Setup the handle */
    if (!(handle = htext_init()))
        return dav_shared_new_error(src->info->request, NULL,
                HTTP_INTERNAL_SERVER_ERROR, "Not enough memory");

    open_params.fsize = src->info->fsize;
    open_params.context = src->info->ctx;
    open_params.extra = src->info->loc.chunks[0].url.query;

    htext_setopt(handle, HTEXTOP_SOURCEURL, src->info->loc.chunks[0].url.path);
    htext_setopt(handle, HTEXTOP_IO_HANDLER, &dav_disk_io_handler);
    htext_setopt(handle, HTEXTOP_IO_HANDLER_DATA, &open_params);
    htext_setopt(handle, HTEXTOP_DESTINATIONURL, remote);

    htext_setopt(handle, HTEXTOP_NUMBEROFSTREAMS, nstreams);

    if (uproxy) {
        htext_setopt(handle, HTEXTOP_USERCERTIFICATE, uproxy);
        htext_setopt(handle, HTEXTOP_USERPRIVKEY, uproxy);
        htext_setopt(handle, HTEXTOP_CAFILE, uproxy);
    }

    htext_setopt(handle, HTEXTOP_CAPATH, "/etc/grid-security/certificates");
    htext_setopt(handle, HTEXTOP_VERIFYPEER, 1);

    htext_setopt(handle, HTEXTOP_LOGCALLBACK, dav_disk_copy_log);
    htext_setopt(handle, HTEXTOP_LOGCALLBACK_DATA, &ddc);
    htext_setopt(handle, HTEXTOP_VERBOSITY, 1);

    ddc.output = src->info->request->output_filters;
    ddc.brigade = apr_brigade_create(src->pool, ddc.output->c->bucket_alloc);
    ddc.request = src->info->request;
    ddc.remote = remote;

    /* Run */
    if (htext_perform(handle) != 0) {
        error = dav_shared_new_error(src->info->request, NULL,
                HTTP_INTERNAL_SERVER_ERROR, "Could not perform the action: %s",
                htext_error_string(handle));
        htext_destroy(handle);
        return error;
    }

    src->info->request->status = 0;
    bkt = apr_bucket_flush_create(ddc.output->c->bucket_alloc);

    /* Poll and feedback */
    error = dav_disk_pool_and_feedback(handle, &ddc);

    /* Finish */
    htext_destroy(handle);

    if (!error) {
        bkt = apr_bucket_eos_create(ddc.output->c->bucket_alloc);
        APR_BRIGADE_INSERT_TAIL(ddc.brigade, bkt);
        if (ap_pass_brigade(ddc.output, ddc.brigade) != APR_SUCCESS)
            error = dav_shared_new_error(src->info->request, NULL,
                    HTTP_INTERNAL_SERVER_ERROR,
                    "Could not write EOS to filter.");
    }

    return error;
}
